package com.sun.xml.internal.bind.v2.util;

import java.util.AbstractSet;
import java.util.NoSuchElementException;
import java.util.HashSet;
import java.util.Collection;
import java.util.Iterator;
import com.sun.xml.internal.bind.v2.runtime.Name;
import javax.xml.namespace.QName;
import java.util.Set;

public final class QNameMap<V>
{
    private static final int DEFAULT_INITIAL_CAPACITY = 16;
    private static final int MAXIMUM_CAPACITY = 1073741824;
    transient QNameMap.Entry<V>[] table;
    transient int size;
    private int threshold;
    private static final float DEFAULT_LOAD_FACTOR = 0.75f;
    private Set<QNameMap.Entry<V>> entrySet;
    
    public QNameMap() {
        this.table = new QNameMap.Entry[16];
        this.entrySet = null;
        this.threshold = 12;
        this.table = new QNameMap.Entry[16];
    }
    
    public void put(String namespaceUri, String localname, V value) {
        assert localname != null;
        assert namespaceUri != null;
        assert localname == localname.intern();
        assert namespaceUri == namespaceUri.intern();
        int hash = hash(localname);
        int i = indexFor(hash, this.table.length);
        QNameMap.Entry<V> e = this.table[i];
        while (e != null) {
            if (e.hash == hash && localname == e.localName && namespaceUri == e.nsUri) {
                e.value = value;
                return;
            }
            e = e.next;
        }
        this.addEntry(hash, namespaceUri, localname, value, i);
        return;
    }
    
    public void put(QName name, V value) {
        this.put(name.getNamespaceURI(), name.getLocalPart(), value);
    }
    
    public void put(Name name, V value) {
        this.put(name.nsUri, name.localName, value);
    }
    
    public V get(String nsUri, String localPart) {
        QNameMap.Entry<V> e = this.getEntry(nsUri, localPart);
        if (e == null) {
            return null;
        }
        return e.value;
    }
    
    public V get(QName name) {
        return this.get(name.getNamespaceURI(), name.getLocalPart());
    }
    
    public int size() {
        return this.size;
    }
    
    public QNameMap<V> putAll(QNameMap<? extends V> map) {
        int numKeysToBeAdded = map.size();
        if (numKeysToBeAdded == 0) {
            return this;
        }
        if (numKeysToBeAdded > this.threshold) {
            int targetCapacity = numKeysToBeAdded;
            if (targetCapacity > 1073741824) {
                targetCapacity = 1073741824;
            }
            int newCapacity;
            for (newCapacity = this.table.length; newCapacity < targetCapacity; newCapacity <<= 1) {}
            if (newCapacity > this.table.length) {
                this.resize(newCapacity);
            }
        }
        for (QNameMap.Entry<? extends V> e : map.entrySet()) {
            this.put(e.nsUri, e.localName, e.getValue());
        }
        return this;
    }
    
    private static int hash(String x) {
        int h = x.hashCode();
        h += ~(h << 9);
        h ^= h >>> 14;
        h += h << 4;
        h ^= h >>> 10;
        return h;
    }
    
    private static int indexFor(int h, int length) {
        return h & length - 1;
    }
    
    private void addEntry(int hash, String nsUri, String localName, V value, int bucketIndex) {
        QNameMap.Entry<V> e = this.table[bucketIndex];
        this.table[bucketIndex] = new QNameMap.Entry(hash, nsUri, localName, value, e);
        if (this.size++ >= this.threshold) {
            this.resize(2 * this.table.length);
        }
    }
    
    private void resize(int newCapacity) {
        QNameMap.Entry[] oldTable = this.table;
        int oldCapacity = oldTable.length;
        if (oldCapacity == 1073741824) {
            this.threshold = Integer.MAX_VALUE;
            return;
        }
        QNameMap.Entry[] newTable = new QNameMap.Entry[newCapacity];
        this.transfer(newTable);
        this.table = newTable;
        this.threshold = newCapacity;
        return;
    }
    
    private void transfer(QNameMap.Entry<V>[] newTable) {
        QNameMap.Entry<V>[] src = this.table;
        int newCapacity = newTable.length;
        for (int j = 0; j < src.length; ++j) {
            QNameMap.Entry<V> e = src[j];
            if (e != null) {
                src[j] = null;
                do {
                    QNameMap.Entry<V> next = e.next;
                    int i = indexFor(e.hash, newCapacity);
                    e.next = newTable[i];
                    newTable[i] = e;
                    e = next;
                } while (e != null);
            }
        }
    }
    
    public QNameMap.Entry<V> getOne() {
        QNameMap.Entry[] table = this.table;
        int length = table.length;
        int i = 0;
        while (i < length) {
            QNameMap.Entry<V> e = table[i];
            if (e != null) {
                return e;
            }
            ++i;
        }
        return null;
    }
    
    public Collection<QName> keySet() {
        Set<QName> r = new HashSet();
        for (QNameMap.Entry<V> e : this.entrySet()) {
            r.add(e.createQName());
        }
        return r;
    }
    
    public boolean containsKey(String nsUri, String localName) {
        return this.getEntry(nsUri, localName) != null;
    }
    
    public boolean isEmpty() {
        return this.size == 0;
    }
    
    public Set<QNameMap.Entry<V>> entrySet() {
        Set<QNameMap.Entry<V>> es = this.entrySet;
        return (es != null) ? es : (this.entrySet = new QNameMap.EntrySet());
    }
    
    private Iterator<QNameMap.Entry<V>> newEntryIterator() {
        return new QNameMap.EntryIterator();
    }
    
    private QNameMap.Entry<V> getEntry(String nsUri, String localName) {
        assert nsUri == nsUri.intern();
        assert localName == localName.intern();
        int hash = hash(localName);
        int i = indexFor(hash, this.table.length);
        QNameMap.Entry<V> e;
        for (e = this.table[i]; e != null && (localName != e.localName || nsUri != e.nsUri); e = e.next) {}
        return e;
    }
    
    @Override
    public String toString() {
        StringBuilder buf = new StringBuilder();
        buf.append('{');
        for (QNameMap.Entry<V> e : this.entrySet()) {
            if (buf.length() > 1) {
                buf.append(',');
            }
            buf.append('[');
            buf.append(e);
            buf.append(']');
        }
        buf.append('}');
        return buf.toString();
    }
    
    static Iterator access$200(QNameMap x0) {
        return x0.newEntryIterator();
    }
    
    static QNameMap.Entry access$300(QNameMap x0, String x1, String x2) {
        return x0.getEntry(x1, x2);
    }
    
    private abstract class HashIterator<E> implements Iterator<E>
    {
        QNameMap.Entry<V> next;
        int index;
        
        HashIterator() {
            QNameMap.Entry<V>[] t = QNameMap.this.table;
            int i = t.length;
            QNameMap.Entry<V> n = null;
            if (QNameMap.this.size != 0) {
                while (i > 0 && (n = t[--i]) == null) {}
            }
            this.next = n;
            this.index = i;
        }
        
        @Override
        public boolean hasNext() {
            return this.next != null;
        }
        
        QNameMap.Entry<V> nextEntry() {
            QNameMap.Entry<V> e = this.next;
            if (e == null) {
                throw new NoSuchElementException();
            }
            QNameMap.Entry<V> n;
            QNameMap.Entry<V>[] t;
            int i;
            for (n = e.next, t = QNameMap.this.table, i = this.index; n == null && i > 0; n = t[--i]) {}
            this.index = i;
            this.next = n;
            return e;
        }
        
        @Override
        public void remove() {
            throw new UnsupportedOperationException();
        }
    }
    
    public static final class Entry<V>
    {
        public final String nsUri;
        public final String localName;
        V value;
        final int hash;
        QNameMap.Entry<V> next;
        
        Entry(int h, String nsUri, String localName, V v, QNameMap.Entry<V> n) {
            this.value = v;
            this.next = n;
            this.nsUri = nsUri;
            this.localName = localName;
            this.hash = h;
        }
        
        public QName createQName() {
            return new QName(this.nsUri, this.localName);
        }
        
        public V getValue() {
            return this.value;
        }
        
        public V setValue(V newValue) {
            V oldValue = this.value;
            this.value = newValue;
            return oldValue;
        }
        
        @Override
        public boolean equals(Object o) {
            if (!(o instanceof QNameMap.Entry)) {
                return false;
            }
            QNameMap.Entry<Object> e = (QNameMap.Entry)o;
            String k1 = this.nsUri;
            String k2 = e.nsUri;
            String k3 = this.localName;
            String k4 = e.localName;
            if (k1 == k2 || (k1 != null && k1.equals(k2) && (k3 == k4 || (k3 != null && k3.equals(k4))))) {
                Object v1 = this.getValue();
                Object v2 = e.getValue();
                if (v1 == v2 || (v1 != null && v1.equals(v2))) {
                    return true;
                }
            }
            return false;
        }
        
        @Override
        public int hashCode() {
            return this.localName.hashCode() ^ ((this.value == null) ? 0 : this.value.hashCode());
        }
        
        @Override
        public String toString() {
            return 34 + this.nsUri + "\",\"" + this.localName + "\"=" + this.getValue();
        }
    }
    
    private class EntryIterator extends QNameMap.HashIterator<QNameMap.Entry<V>>
    {
        @Override
        public QNameMap.Entry<V> next() {
            return this.nextEntry();
        }
    }
    
    private class EntrySet extends AbstractSet<QNameMap.Entry<V>>
    {
        @Override
        public Iterator<QNameMap.Entry<V>> iterator() {
            return QNameMap.access$200(QNameMap.this);
        }
        
        @Override
        public boolean contains(Object o) {
            if (!(o instanceof QNameMap.Entry)) {
                return false;
            }
            QNameMap.Entry<V> e = (QNameMap.Entry)o;
            QNameMap.Entry<V> candidate = QNameMap.access$300(QNameMap.this, e.nsUri, e.localName);
            return candidate != null && candidate.equals(e);
        }
        
        @Override
        public boolean remove(Object o) {
            throw new UnsupportedOperationException();
        }
        
        @Override
        public int size() {
            return QNameMap.this.size;
        }
    }
}
